"""Constants for the Derivative integration."""

DOMAIN = "derivative"

CONF_ROUND_DIGITS = "round"
CONF_TIME_WINDOW = "time_window"
CONF_UNIT = "unit"
CONF_UNIT_PREFIX = "unit_prefix"
CONF_UNIT_TIME = "unit_time"
