"""Constants for the Tautulli integration."""
from logging import Logger, getLogger

ATTR_TOP_USER = "top_user"

CONF_MONITORED_USERS = "monitored_users"
DEFAULT_NAME = "Tautulli"
DOMAIN = "tautulli"
LOGGER: Logger = getLogger(__package__)
