"""TotalConnect constants."""

DOMAIN = "totalconnect"
CONF_USERCODES = "usercodes"
CONF_LOCATION = "location"
AUTO_BYPASS = "auto_bypass_low_battery"

# Most TotalConnect alarms will work passing '-1' as usercode
DEFAULT_USERCODE = "-1"
