"""Constants for the tankerkoenig integration."""

DOMAIN = "tankerkoenig"
NAME = "tankerkoenig"

CONF_FUEL_TYPES = "fuel_types"
CONF_STATIONS = "stations"

DEFAULT_RADIUS = 2
DEFAULT_SCAN_INTERVAL = 30

FUEL_TYPES = {"e5": "Super", "e10": "Super E10", "diesel": "Diesel"}

ATTR_BRAND = "brand"
ATTR_CITY = "city"
ATTR_FUEL_TYPE = "fuel_type"
ATTR_HOUSE_NUMBER = "house_number"
ATTR_POSTCODE = "postcode"
ATTR_STATION_NAME = "station_name"
ATTR_STREET = "street"
ATTRIBUTION = "Data provided by https://www.tankerkoenig.de"
