"""Constants for the IntelliFire integration."""
from __future__ import annotations

import logging

DOMAIN = "intellifire"

CONF_USER_ID = "user_id"

LOGGER = logging.getLogger(__package__)

CONF_SERIAL = "serial"
